import os, sys, threading
from queue import Queue
from time import sleep
if sys.platform.startswith('linux'):
    def set_nonblocking(fd):
        import fcntl
        flag = fcntl.fcntl(fd, fcntl.F_GETFL)
        fcntl.fcntl(fd, fcntl.F_SETFL, flag | os.O_NONBLOCK)
else:
    def set_nonblocking(fd):
        import msvcrt
        from ctypes import windll, byref, wintypes, WinError, POINTER
        from ctypes.wintypes import HANDLE, DWORD, BOOL
        PIPE_NOWAIT = DWORD(0x00000001)
        def pipe_no_wait(pipefd):
            SetNamedPipeHandleState = windll.kernel32.SetNamedPipeHandleState
            SetNamedPipeHandleState.argtypes = [HANDLE, POINTER(DWORD), POINTER(DWORD), POINTER(DWORD)]
            SetNamedPipeHandleState.restype = BOOL
            h = msvcrt.get_osfhandle(pipefd)
            res = windll.kernel32.SetNamedPipeHandleState(h, byref(PIPE_NOWAIT), None, None)
            if res == 0:
                print(WinError())
                return False
            return True
        return pipe_no_wait(fd)
    

class AsyncFileReader:
    def __init__(self, f):
        self.f = f
        self.queue = Queue()
        self.thread = threading.Thread(target=self._read_thread)
        self.keep_polling = True
        self.thread.daemon = True
        self.thread.start()
    
    def _read_thread(self):
        blocking = True
        if sys.platform.startswith('linux'):
            set_nonblocking(self.f.fileno())
            blocking = False
        try:
            while self.keep_polling and not self.f.closed:
                while not self.f.closed:
                    line = self.f.readline()
                    if not line: break
                    self.queue.put(line)
                if not blocking:
                    sleep(0.015)
        except:
            return

    def available(self):
        return not self.queue.empty()

    def readline(self):
        if self.available():
            return self.queue.get()
        return ''

    def print(self):
        while self.available():
            print(self.readline(), flush=True, end='')

    def stop(self):
        self.keep_polling = False
        self.thread.join()
        self.print()




import os, concurrent.futures
from .build_target import BuildTarget
from .build_dependency import BuildDependency
from .util import save_file_if_contents_changed
from .system import console


def get_cmake_path_list(paths):
    pathlist = '' 
    for path in paths: pathlist += f'\n    "{path}"'
    return pathlist


def get_flattened_deps(root: BuildDependency):
    # deps have to be sorted in [parent] [child] order for Unix linkers
    ordered = []
    def add_unique_items(deps):
        for child in deps:
            if child in ordered: # already in deps list, so we need to move it lower
                ordered.remove(child)
            ordered.append(child)
            add_unique_items(child.children)
    add_unique_items(root.children)

    if len(ordered) > 1:
        names = [dep.name for dep in ordered]
        console(f'  - Dependency Order:\n        {" ".join(names)}')
    return ordered


def save_dependencies_cmake(root: BuildDependency):
    if not root.build_dir_exists():
        return # probably CLEAN, so nothing to save
    outfile = f'{root.build_dir}/mama-dependencies.cmake'
    if not root.children:
        if os.path.exists(outfile):
            os.remove(outfile) # no more deps, get rid of the dependency file
        return
    text = '''
# This file is auto-generated by mama build. Do not modify by hand!
'''
    flatdeps = get_flattened_deps(root)
    for dep in flatdeps:
        includes  = get_cmake_path_list(dep.target.exported_includes)
        libraries = get_cmake_path_list(dep.target.exported_libs)
        text += f'''
# Package {dep.name}
set({dep.name}_INCLUDES {includes})
set({dep.name}_LIBS {libraries})
set(MAMA_INCLUDES ${{MAMA_INCLUDES}} ${{{dep.name}_INCLUDES}})
set(MAMA_LIBS     ${{MAMA_LIBS}}     ${{{dep.name}_LIBS}})
'''
    save_file_if_contents_changed(outfile, text)


def get_mama_dependencies_cmake(root: BuildDependency, build:str):
    if not root.children:
        return ''
    return f'''# get MAMA_INCLUDES and MAMA_LIBS for this platform; verbose for CLion CMake parser
    include("{root.dep_dir}/{build}/mama-dependencies.cmake")'''


def mama_cmake_path(root: BuildDependency):
    return f'{root.src_dir}/mama.cmake'


def save_mama_cmake(root: BuildDependency):
    save_dependencies_cmake(root)
    # note: we save verbose include directives, because CLion has a hard time detecting macro paths
    text = f'''
# This file is auto-generated by mama build. Do not modify by hand!
if(CMAKE_CXX_COMPILER_ID MATCHES "Clang")
    set(CLANG TRUE)
elseif(CMAKE_CXX_COMPILER_ID MATCHES "GNU")
    set(GCC TRUE)
endif()
set(MAMA_INCLUDE "")
set(MAMA_LIBS "")
if(ANDROID OR ANDROID_NDK)
    set(MAMA_BUILD "android")
    {get_mama_dependencies_cmake(root, 'android')}
elseif(WIN32)
    set(MAMA_BUILD "windows")
    {get_mama_dependencies_cmake(root, 'windows')}
elseif(APPLE)
  if(IOS_PLATFORM)
    set(IOS TRUE)
    set(MAMA_BUILD "ios")
    {get_mama_dependencies_cmake(root, 'ios')}
  else()
    set(MACOS TRUE)
    set(MAMA_BUILD "macos")
    {get_mama_dependencies_cmake(root, 'macos')}
  endif()
elseif(UNIX)
    set(LINUX TRUE)
    set(MAMA_BUILD "linux")
    {get_mama_dependencies_cmake(root, 'linux')}
else()
    message(FATAL_ERROR "mama build: Unsupported Platform!")
    set(MAMA_BUILD "???")
endif()
'''
    save_file_if_contents_changed(mama_cmake_path(root), text)


def _load_child_dependencies(root: BuildDependency, parallel=True):
    changed = False
    if parallel:
        futures = []
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as e:
            for dep in root.children:
                futures.append(e.submit(_load_dependency_chain, dep))
        for f in futures:
            changed |= f.result()
    else:
        for dep in root.children:
            changed |= _load_dependency_chain(dep)
    return changed


def _load_dependency_chain(dep: BuildDependency):
    if dep.already_loaded:
        return dep.should_rebuild

    changed = dep.load()
    changed |= _load_child_dependencies(dep)
    return changed


def load_dependency_chain(root: BuildDependency):
    with concurrent.futures.ThreadPoolExecutor() as e:
        def load_dependency(dep: BuildDependency):
            if dep.already_loaded:
                return dep.should_rebuild
            changed = dep.load()
            futures = []
            for child in dep.children:
                futures.append(e.submit(load_dependency, child))
            for f in futures:
                changed |= f.result()
            dep.after_load()
            return changed
        load_dependency(root)


def execute_task_chain(root: BuildDependency):
    if not os.path.exists(mama_cmake_path(root)):
        save_mama_cmake(root) # save a dummy mama.cmake before build

    for dep in root.children:
        execute_task_chain(dep)
    
    save_mama_cmake(root)
    root.target.execute_tasks()


def find_dependency(root: BuildDependency, name) -> BuildDependency:
    if root.name == name:
        return root
    for dep in root.children:
        found = find_dependency(dep, name)
        if found: return found
    return None



#!/usr/bin/python3.6
import sys, os, importlib
from .system import console, execute
from .util import glob_with_extensions, glob_folders_with_name_match
from .build_config import BuildConfig
from .build_target import BuildTarget
from .build_dependency import BuildDependency
from .dependency_chain import load_dependency_chain, execute_task_chain, find_dependency
from .init_project import mama_init_project

def print_usage():
    console('mama [actions...] [args...]')
    console('  actions:')
    console('    init       - create initial mamafile.py and CMakeLists.txt')
    console('    build      - update, configure and build main project or specific target')
    console('    clean      - clean main project or specific target')
    console('    rebuild    - clean, update, configure and build main project or specific target')
    console('    configure  - run CMake configuration on main project or specific target')
    console('    update     - update specific target dependency by calling git pull')
    console('    reclone    - wipe specific target dependency and clone it again')
    console('    test       - run tests for main project or specific target')
    console('    add        - add new dependency')
    console('    new        - create new mama build file')
    console('    open=<tgt> - open a project file')
    console('    help       - shows this help list')
    console('  args:')
    console('    windows    - build for windows')
    console('    linux      - build for linux')
    console('    macos      - build for macos')
    console('    ios        - build for ios')
    console('    android    - build for android')
    console('    clang      - prefer clang for linux (default on linux/macos/ios/android)')
    console('    gcc        - prefer gcc for linux')
    console('    fortran    - enable automatic fortran detection (or configure this in mamafile)')
    console('    release    - (default) CMake configuration RelWithDebInfo')
    console('    debug      - CMake configuration Debug')
    console('    jobs=N     - Max number of parallel compilations. (default=system.core.count)')
    console('    target=P   - Name of the target')
    console('    all        - Short for target=all')
    console('  examples:')
    console('    mama init                     Initialize a new project. Tries to create mamafile.py and CMakeLists.txt')
    console('    mama build                    Update and build main project only.')
    console('    mama clean                    Cleans main project only.')
    console('    mama clean all                Cleans EVERYTHING in the dependency chain.')
    console('    mama rebuild                  Cleans, update and build main project only.')
    console('    mama update build             Update all dependencies and then build.')
    console('    mama build target=dep1        Update and build dep1 only.')
    console('    mama configure                Run CMake configuration on main project only.')
    console('    mama configure target=all     Run CMake configuration on main project and all deps.')
    console('    mama configure all            Run CMake configuration on main project and all deps.')
    console('    mama reclone target=dep1      Wipe target dependency completely and clone again.')
    console('    mama test                     Run tests on main project.')
    console('    mama test=arg                 Run tests on main project with an argument.')
    console('    mama test="arg1 arg2"         Run tests on main project with multiple arguments.')
    console('    mama test target=dep1         Run tests on target dependency project.')
    console('  environment:')
    console('    setenv("NINJA")               Path to NINJA build executable')
    console('    setenv("ANDROID_HOME")        Path to Android SDK if auto-detect fails')


def open_project(config: BuildConfig, root_dependency: BuildDependency):
    name = config.target if config.target and config.target != 'all' else config.open
    found = root_dependency if name == 'root' else find_dependency(root_dependency, name)
    if not found:
        raise KeyError(f'No project named {name}')
    
    if config.windows:
        solutions = glob_with_extensions(found.build_dir, ['.sln'])
        if not solutions:
            raise EnvironmentError('Could not find any Visual Studio solutions!')
        execute(f'start {solutions[0]}', echo=True)

    elif config.macos or config.ios:
        projects = glob_folders_with_name_match(found.build_dir, ['.xcodeproj'])
        if not projects:
            raise EnvironmentError('Could not find any Xcode projects!')
        execute(f'open {projects[0]}', echo=True)

    elif config.linux:
        raise EnvironmentError('Linux IDE selection not implemented. Try opening this folder with CLion.')
        #execute(f'xdg-open', echo=True)

    elif config.android:
        raise EnvironmentError('Android IDE selection not implemented. Try opening this folder with Android Studio.')


def main():
    console(f'========= Mama Build Tool ==========')
    if sys.version_info < (3, 6):
        console('FATAL ERROR: MamaBuild requires Python 3.6')
        exit(-1)

    if len(sys.argv) == 1 or 'help' in sys.argv:
        print_usage()
        exit(-1)

    config = BuildConfig(sys.argv[1:])
    source_dir = os.getcwd()
    name = os.path.basename(source_dir)
    root = BuildDependency(name, config, BuildTarget, src=source_dir, is_root=True)

    if config.mama_init:
        mama_init_project(root)
        return

    has_cmake = root.cmakelists_exists()
    if not root.mamafile_exists() and not has_cmake:
        console('FATAL ERROR: mamafile.py not found')
        exit(-1)

    ## TODO Remove this eventually when we support non-cmake root builds
    if not has_cmake:
        console('FATAL ERROR: CMakeLists.txt not found')
        exit(-1)

    if config.update:
        if not config.target:
            config.target = 'all'
            console(f'Updating all targets')
        else:
            console(f'Updating {config.target} target')

    if config.rebuild:
        config.build = True
        config.clean = True

    if config.clean and not config.target:
        root.clean()

    load_dependency_chain(root)
    if config.android: config.init_ndk_path()
    if config.raspi:   config.init_raspi_path()
    execute_task_chain(root)

    if config.open:
        open_project(config, root)


def __main__():
    main()


if __name__ == '__main__':
    main()

